"""
Sub-module for adding observational data to plots.

Includes the ObservationalData object and helper functions
to convert data to this new format.
"""

from velociraptor.observations.objects import ObservationalData


def load_observation(filename: str):
    """
    Load an observation from file filename. This should be in the
    standard velociraptor format.

    Parameters
    ----------

    filename: str
        Filename of the observational dataset that you wish to load.
        Should probably end in .hdf5. See the documentation for
        :class:`velociraptor.observations.objects.ObservationalData`
        for more information.

    Returns
    -------

    velociraptor.observations.objects.ObservationalData:
        Observational data instance read from file.
    """

    data = ObservationalData()
    data.load(filename)

    return data



"""
Main objects for holding information relating to the autoplotter.
"""

from velociraptor import VelociraptorCatalogue
from velociraptor.autoplotter.lines import VelociraptorLine, valid_line_types
from velociraptor.exceptions import AutoPlotterError
from velociraptor.observations.objects import ObservationalData

import velociraptor.autoplotter.plot as plot

from unyt import unyt_quantity, unyt_array, matplotlib_support
from unyt.exceptions import UnitConversionError
from numpy import log10, linspace, logspace, array, logical_and
from matplotlib.pyplot import Axes, Figure, close
from yaml import safe_load
from typing import Union, List, Dict, Tuple

from os import path, mkdir
from functools import reduce
from collections import OrderedDict

valid_plot_types = [
    "scatter",
    "2dhistogram",
    "massfunction",
    "histogram",
    "adaptivemassfunction",
]

matplotlib_support.label_style = "[]"


class VelociraptorPlot(object):
    """
    Object representing a single figure of x against y.
    """

    # Forward declarations
    # What type of plot are we?
    plot_type: str
    # variable to plot on the x-axis
    x: str
    # variable to plot on the y-axis
    y: str
    # log the x/y axes?
    x_log: bool
    y_log: bool
    # Units for x/y
    x_units: unyt_quantity
    y_units: unyt_quantity
    # Plot limits for x/y
    x_lim: List[Union[unyt_quantity, None]]
    y_lim: List[Union[unyt_quantity, None]]
    # Shading for x/y
    x_shade: List[Union[unyt_quantity, None]]
    y_shade: List[Union[unyt_quantity, None]]
    # Labels for x and y axes (no units)
    x_label: Union[None, str] = None
    y_label: Union[None, str] = None
    x_label_override: Union[None, str] = None
    y_label_override: Union[None, str] = None
    comment: Union[None, str]
    # plot median/mean line and give it properties
    mean_line: Union[None, VelociraptorLine]
    median_line: Union[None, VelociraptorLine]
    mass_function_line: Union[None, VelociraptorLine]
    adaptive_mass_function_line: Union[None, VelociraptorLine]
    histogram_line: Union[None, VelociraptorLine]
    # Binning for x, y axes.
    number_of_bins: int
    x_bins: unyt_array
    y_bins: unyt_array
    # Select a specific strucutre type?
    select_structure_type: Union[None, int]
    structure_mask: Union[None, array]
    selection_mask: Union[None, array]
    # Where should the legend and z, a information be placed?
    legend_loc: str
    redshift_loc: str
    comment_loc: str
    # Observational data
    observational_data: List[ObservationalData]
    observational_data_redshift_bracketing: List[List[float]]
    observational_data_directory: str

    def __init__(
        self,
        filename: str,
        data: Dict[str, Union[Dict, str]],
        observational_data_directory: str,
    ):
        """
        Initialise the plot object variables.
        """
        self.filename = filename
        self.data = data
        self.observational_data_directory = observational_data_directory

        self._parse_data()

        return

    def _parse_coordinate_quantity(self, coordinate: str) -> None:
        """
        Parses x or y to self.{x,y} and self.{x,y}_units.
        """
        try:
            setattr(self, coordinate, self.data[coordinate]["quantity"])
            setattr(
                self,
                f"{coordinate}_units",
                unyt_quantity(1.0, units=self.data[coordinate]["units"]),
            )
        except KeyError:
            raise AutoPlotterError(
                f"You must provide an {coordinate}-quantity and units to plot for {self.filename}"
            )

        return

    def _parse_coordinate_quantity_units(self, coordinate: str) -> None:
        """
        Parses x or y to self.{x,y}_units.
        """

        try:
            units = unyt_quantity(
                1.0, self.data[coordinate].get("units", "dimensionless")
            )
        except KeyError:
            units = unyt_quantity(1.0, units="dimensionless")

        setattr(self, f"{coordinate}_units", units)

        return

    def _set_coordinate_quantity_none(self, coordinate: str) -> None:
        """
        Sets a coordinates quantity and units to None and (dimensionless)
        respectively. Useful for e.g. mass functions or histograms.
        """

        setattr(self, coordinate, None)
        setattr(self, f"{coordinate}_units", unyt_quantity(1.0, units=None))

        return

    def _parse_coordinate_limit(self, coordinate: str) -> None:
        """
        Parses the x or y limit to {x,y}_limit.
        """
        setattr(self, f"{coordinate}_lim", [None, None])

        try:
            getattr(self, f"{coordinate}_lim")[0] = unyt_quantity(
                float(self.data[coordinate]["start"]),
                units=self.data[coordinate]["units"],
            )
        except KeyError:
            pass

        try:
            getattr(self, f"{coordinate}_lim")[1] = unyt_quantity(
                float(self.data[coordinate]["end"]),
                units=self.data[coordinate]["units"],
            )
        except KeyError:
            pass

        return

    def _parse_coordinate_shade(self, coordinate: str) -> None:
        """
        Parses x and y's shade: below, above to a list.
        """
        setattr(self, f"{coordinate}_shade", [None, None])

        try:
            getattr(self, f"{coordinate}_shade")[0] = unyt_quantity(
                float(self.data[coordinate]["shade"]["below"]),
                units=self.data[coordinate]["units"],
            )
        except KeyError:
            pass

        try:
            getattr(self, f"{coordinate}_shade")[1] = unyt_quantity(
                float(self.data[coordinate]["shade"]["above"]),
                units=self.data[coordinate]["units"],
            )
        except KeyError:
            pass

    def _parse_coordinate_log(self, coordinate: str) -> None:
        """
        Parses x_log from the parameter file data.
        """

        try:
            setattr(self, f"{coordinate}_log", bool(self.data[coordinate]["log"]))
        except KeyError:
            setattr(self, f"{coordinate}_log", True)

        return

    def _parse_coordinate_label_override(self, coordinate: str) -> None:
        """
        Parses {x,y}_label_override.
        """

        try:
            setattr(
                self,
                f"{coordinate}_label_override",
                self.data[coordinate]["label_override"],
            )
        except KeyError:
            setattr(self, f"{coordinate}_label_override", None)

        return

    def _parse_line(self, line_type: str) -> None:
        """
        Parses a line type to {line_type}_line.
        """

        try:
            setattr(
                self,
                f"{line_type}_line",
                VelociraptorLine(line_type, self.data[line_type]),
            )
        except KeyError:
            setattr(self, f"{line_type}_line", None)

        return

    def _parse_lines(self) -> None:
        """
        Parses all lines to VelociraptorLine objects using _parse_line individually.
        """

        for line_type in valid_line_types:
            self._parse_line(line_type)

        return

    def _parse_structure_type(self) -> None:
        """
        Parses the structure type selector to select_structure_type, as well as
        creating the structure_mask
        """
        try:
            self.select_structure_type = int(self.data["select_structure_type"])
        except KeyError:
            self.select_structure_type = None

        # Initialise to None; we set/create this when we first have access to the
        # catalogue in the plotting functions.
        self.structure_mask = None

        return

    def _parse_selection_mask(self) -> None:
        """
        Parses the selection mask selector.
        """

        try:
            self.selection_mask = self.data["selection_mask"]
        except KeyError:
            self.selection_mask = None

        return

    def _parse_number_of_bins(self) -> None:
        """
        Parses the number of bins.
        """

        try:
            self.number_of_bins = int(self.data["number_of_bins"])
        except KeyError:
            self.number_of_bins = 128

        return

    def _parse_comment(self) -> None:
        """
        Parse the extra text comment.
        """

        try:
            self.comment = str(self.data["comment"])
        except KeyError:
            self.comment = None

        return

    def _parse_loc(self) -> None:
        """
        Parses the legend and redshift location.
        """

        valid_locs = [
            "upper right",
            "upper left",
            "lower left",
            "lower right",
            "right",
            "lower center",
            "upper center",
            "center",
        ]

        try:
            self.legend_loc = str(self.data["legend_loc"])
            if self.legend_loc not in valid_locs:
                raise AutoPlotterError(
                    f"Choice of legend_loc {self.legend_loc} invalid. "
                    f"Choose from one of {valid_locs}"
                )
        except KeyError:
            self.legend_loc = "lower left"

        try:
            self.redshift_loc = str(self.data["redshift_loc"])
            if self.redshift_loc not in valid_locs:
                raise AutoPlotterError(
                    f"Choice of redshift_loc {self.redshift_loc} invalid. "
                    f"Choose from one of {valid_locs}"
                )
        except KeyError:
            # Set redshift based on legend. Note that repeated calls to
            # `replace` would simply overwrite each other.

            replacements = OrderedDict(
                {
                    "upper": "lower",
                    "lower": "upper",
                    "left": "right",
                    "right": "left",
                    "center": "right",
                }
            )

            self.redshift_loc = " ".join(
                [b for a, b in replacements.items() if a in self.legend_loc.split(" ")]
            )

            self.redshift_loc = (
                "lower center" if self.redshift_loc == "left" else self.redshift_loc
            )

        try:
            self.comment_loc = str(self.data["comment_loc"])
            if self.comment_loc not in valid_locs:
                raise AutoPlotterError(
                    f"Choice of comment_loc {self.comment_loc} invalid. "
                    f"Choose from one of {valid_locs}"
                )
        except KeyError:
            # Set comment loc based on redshift label. Need to make sure that this does
            # not overlap with the legend too!

            replacements = OrderedDict(
                {
                    "upper": "upper",
                    "lower": "lower",
                    "left": "right",
                    "right": "left",
                    "center": "center",
                }
            )

            self.comment_loc = " ".join(
                [
                    b
                    for a, b in replacements.items()
                    if a in self.redshift_loc.split(" ")
                ]
            )

            self.comment_loc = (
                "upper center" if self.comment_loc == "left" else self.comment_loc
            )

        return

    def _parse_coordinate_histogram_bin(self, coordinate: str) -> None:
        """
        Parses the histogram bins for a given histogram axis, given by
        co-ordinate. Specifically x:start and x:end.
        """

        start, end = getattr(self, f"{coordinate}_lim")

        if getattr(self, f"{coordinate}_log"):
            # Need to strip units, unfortunately
            setattr(
                self,
                f"{coordinate}_bins",
                unyt_array(
                    logspace(log10(start), log10(end), self.number_of_bins + 1),
                    units=start.units,
                ),
            )
        else:
            # Can get away with this one without stripping
            setattr(
                self,
                f"{coordinate}_bins",
                linspace(start, end, self.number_of_bins + 1),
            )

        return

    def _parse_scatter(self) -> None:
        """
        Parses the required variables for producing a scatter plot.
        """

        for coordinate in ["x", "y"]:
            self._parse_coordinate_quantity(coordinate)
            self._parse_coordinate_log(coordinate)
            self._parse_coordinate_limit(coordinate)
            self._parse_coordinate_label_override(coordinate)
            self._parse_coordinate_shade(coordinate)

        self._parse_loc()
        self._parse_comment()
        self._parse_lines()
        self._parse_structure_type()
        self._parse_selection_mask()

        return

    def _parse_2dhistogram(self) -> None:
        """
        Parses the required variables for producing a background
        2d histogram plot.
        """

        # Requires everything for the scatter, but with extra tacked
        # on.

        self._parse_scatter()
        self._parse_number_of_bins()

        for coordinate in ["x", "y"]:
            self._parse_coordinate_histogram_bin(coordinate)

        return

    def _parse_common_histogramtype(self) -> None:
        """
        Common parsing between histogram and mass function.
        """

        self._parse_coordinate_quantity("x")
        self._set_coordinate_quantity_none("y")
        self._parse_coordinate_quantity_units("y")

        for coordinate in ["x", "y"]:
            self._parse_coordinate_log(coordinate)
            self._parse_coordinate_limit(coordinate)
            self._parse_coordinate_label_override(coordinate)
            self._parse_coordinate_shade(coordinate)

        self._parse_number_of_bins()
        self._parse_coordinate_histogram_bin("x")
        self._parse_loc()
        self._parse_comment()
        self._parse_structure_type()
        self._parse_selection_mask()

        return

    def _parse_massfunction(self) -> None:
        """
        Parses the required variables for producing a mass function
        plot.

        TODO: Re-write the mass function in a better way to be the
        same as other lines.
        """

        self._parse_common_histogramtype()

        # A bit of a hacky workaround - improve this in the future
        # by combining this functionality properly into the
        # VelociraptorLine methods.
        self.mass_function_line = VelociraptorLine(
            line_type="mass_function",
            line_data=dict(
                plot=True,
                log=self.x_log,
                number_of_bins=self.number_of_bins,
                start=dict(value=self.x_lim[0].value, units=self.x_lim[0].units),
                end=dict(value=self.x_lim[1].value, units=self.x_lim[1].units),
            ),
        )

        return

    def _parse_adaptivemassfunction(self) -> None:
        """
        Parses the required variables for producing a mass function
        plot.

        TODO: Re-write the mass function in a better way to be the
        same as other lines.
        """

        self._parse_common_histogramtype()

        # A bit of a hacky workaround - improve this in the future
        # by combining this functionality properly into the
        # VelociraptorLine methods.
        self.adaptive_mass_function_line = VelociraptorLine(
            line_type="adaptive_mass_function",
            line_data=dict(
                plot=True,
                log=self.x_log,
                number_of_bins=self.number_of_bins,
                start=dict(value=self.x_lim[0].value, units=self.x_lim[0].units),
                end=dict(value=self.x_lim[1].value, units=self.x_lim[1].units),
                adaptive=True,
            ),
        )

        return

    def _parse_histogram(self) -> None:
        """
        Parses the required variables for producing a 1D histogram plot.
        """

        # Same as mass function, unsurprisingly!
        self._parse_common_histogramtype()

        self.histogram_line = VelociraptorLine(
            line_type="histogram",
            line_data=dict(
                plot=True,
                log=self.x_log,
                number_of_bins=self.number_of_bins,
                start=dict(value=self.x_lim[0].value, units=self.x_lim[0].units),
                end=dict(value=self.x_lim[1].value, units=self.x_lim[1].units),
            ),
        )

        return

    def _parse_data(self):
        """
        Federates out data parsing to individual functions based on the
        plot type.
        """

        try:
            self.plot_type = self.data["type"]
        except KeyError:
            self.plot_type = "scatter"

        if self.plot_type not in valid_plot_types:
            raise AutoPlotterError(
                f"Plot type {self.plot_type} is not valid. Please choose from {valid_plot_types}."
            )

        getattr(self, f"_parse_{self.plot_type}")()

        self._parse_observational_data()

        return

    def _parse_observational_data(self):
        """
        Parses the observational data segment.
        """

        self.observational_data = []
        self.observational_data_redshift_bracketing = []

        try:
            obs_data = self.data["observational_data"]

            for data in obs_data:
                observational_data_file_path = (
                    f"{self.observational_data_directory}/{data['filename']}"
                )

                if not path.exists(observational_data_file_path):
                    raise AutoPlotterError(
                        f"Unable to find file at {data['filename']}."
                    )
                else:
                    obs_data_instance = ObservationalData()
                    obs_data_instance.load(observational_data_file_path)

                    try:
                        obs_data_instance.x.convert_to_units(self.x_units)
                        obs_data_instance.y.convert_to_units(self.y_units)
                    except UnitConversionError as e:
                        raise AutoPlotterError(
                            f"Unable to convert observational data units for "
                            f"{data['filename']} in plot {self.filename}. See: {e}."
                        )

                    self.observational_data.append(obs_data_instance)

                    minimum_redshift = data.get("minimum_redshift", -1e9)
                    maximum_redshift = data.get("maximum_redshift", 1e9)

                    bracket = [minimum_redshift, maximum_redshift]

                    self.observational_data_redshift_bracketing.append(bracket)

        except KeyError:
            pass

        return

    def _add_shading_to_axes(self, ax: Axes) -> None:
        """
        Adds any required shading.
        """

        common_args = dict(zorder=-10, alpha=0.3, color="grey", linewidth=0)

        # First do horizontal
        if self.x_shade[0] is not None:
            ax.axvspan(self.x_lim[0], self.x_shade[0], **common_args)

        if self.x_shade[1] is not None:
            ax.axvspan(self.x_shade[1], self.x_lim[1], **common_args)

        # Vertical
        if self.y_shade[0] is not None:
            ax.axhspan(self.y_lim[0], self.y_shade[0], **common_args)

        if self.y_shade[1] is not None:
            ax.axhspan(self.y_shade[1], self.y_lim[1], **common_args)

        return

    def _add_lines_to_axes(self, ax: Axes, x: unyt_array, y: unyt_array) -> None:
        """
        Adds any lines present to the given axes.
        """

        if self.median_line is not None:
            self.median_line.plot_line(ax=ax, x=x, y=y, label="Median")
        if self.mean_line is not None:
            self.mean_line.plot_line(ax=ax, x=x, y=y, label="Mean")

        return

    def get_quantity_from_catalogue_with_mask(
        self, quantity: str, catalogue: VelociraptorCatalogue
    ) -> unyt_array:
        """
        Get a quantity from the catalogue using the mask.
        """

        x = reduce(getattr, quantity.split("."), catalogue)
        # We give each dataset a custom name, that gets ruined when masking
        # in versions of unyt less than 2.6.0
        name = x.name

        if self.structure_mask is not None:
            x = x[self.structure_mask]
            x.name = name
        elif self.selection_mask is not None:
            # Create mask
            self.structure_mask = reduce(
                getattr, self.selection_mask.split("."), catalogue
            ).astype(bool)

            if self.select_structure_type is not None:
                self.structure_mask = logical_and(
                    self.structure_mask,
                    catalogue.structure_type.structuretype
                    == self.select_structure_type,
                )

            x = x[self.structure_mask]
            x.name = name
        elif self.select_structure_type is not None:
            # Need to create mask
            self.structure_mask = (
                catalogue.structure_type.structuretype == self.select_structure_type
            )

            x = x[self.structure_mask]
            x.name = name

        return x

    def _make_plot_scatter(
        self, catalogue: VelociraptorCatalogue
    ) -> Tuple[Figure, Axes]:
        """
        Makes a scatter plot and returns the figure and axes.
        """

        x = self.get_quantity_from_catalogue_with_mask(self.x, catalogue)
        x.convert_to_units(self.x_units)
        y = self.get_quantity_from_catalogue_with_mask(self.y, catalogue)
        y.convert_to_units(self.y_units)

        fig, ax = plot.scatter_x_against_y(x, y)
        self._add_lines_to_axes(ax=ax, x=x, y=y)

        return fig, ax

    def _make_plot_2dhistogram(
        self, catalogue: VelociraptorCatalogue
    ) -> Tuple[Figure, Axes]:
        """
        Makes a 2d histogram plot and returns the figure and axes.
        """

        x = self.get_quantity_from_catalogue_with_mask(self.x, catalogue)
        x.convert_to_units(self.x_units)
        y = self.get_quantity_from_catalogue_with_mask(self.y, catalogue)
        y.convert_to_units(self.y_units)

        self.x_bins.convert_to_units(self.x_units)
        self.y_bins.convert_to_units(self.y_units)

        fig, ax = plot.histogram_x_against_y(x, y, self.x_bins, self.y_bins)
        self._add_lines_to_axes(ax=ax, x=x, y=y)

        return fig, ax

    def _make_plot_massfunction(
        self, catalogue: VelociraptorCatalogue
    ) -> Tuple[Figure, Axes]:
        """
        Makes a mass function plot and returns the figure and axes.
        """

        x = self.get_quantity_from_catalogue_with_mask(self.x, catalogue)
        x.convert_to_units(self.x_units)

        # A bit of an odd line but we want to get whichever one is defined
        mass_function_line = getattr(
            self,
            "mass_function_line",
            getattr(self, "adaptive_mass_function_line", None),
        )

        mass_function_line.create_line(
            x=x, y=None, box_volume=catalogue.units.comoving_box_volume
        )

        self.x_bins = mass_function_line.bins

        mass_function_line.output[1].convert_to_units(self.y_units)
        mass_function_line.output[2].convert_to_units(self.y_units)

        fig, ax = plot.mass_function(
            x=x, x_bins=self.x_bins, mass_function=mass_function_line
        )

        return fig, ax

    def _make_plot_adaptivemassfunction(
        self, catalogue: VelociraptorCatalogue
    ) -> Tuple[Figure, Axes]:
        """
        Makes the _adaptive_ mass function plot. Same as mass function.
        """
        return self._make_plot_massfunction(catalogue=catalogue)

    def _make_plot_histogram(
        self, catalogue: VelociraptorCatalogue
    ) -> Tuple[Figure, Axes]:
        """
        Make histogram plot and return the figure and axes.
        """

        x = self.get_quantity_from_catalogue_with_mask(self.x, catalogue)
        x.convert_to_units(self.x_units)

        self.x_bins.convert_to_units(self.x_units)

        self.histogram_line.create_line(
            x=x, y=None, box_volume=catalogue.units.comoving_box_volume
        )

        self.histogram_line.output[1].convert_to_units(self.y_units)

        fig, ax = plot.histogram(x=x, x_bins=self.x_bins, histogram=self.histogram_line)

        return fig, ax

    def make_plot(
        self, catalogue: VelociraptorCatalogue, directory: str, file_extension: str
    ):
        """
        Federates out data parsing to individual functions based on the
        plot type.
        """

        with matplotlib_support:
            fig, ax = getattr(self, f"_make_plot_{self.plot_type}")(catalogue=catalogue)
            if self.x_log:
                ax.set_xscale("log")
            if self.y_log:
                ax.set_yscale("log")

            self._add_shading_to_axes(ax)

            for data, bracket in zip(
                self.observational_data, self.observational_data_redshift_bracketing
            ):
                if bracket[0] < catalogue.z and bracket[1] > catalogue.z:
                    data.plot_on_axes(ax, errorbar_kwargs=dict(zorder=-10))

            try:
                ax.set_xlim(*self.x_lim)
                ax.set_ylim(*self.y_lim)
            except:
                # If we have an empty plot those lines will make us crash!
                # Hopefully, waiting until after the observational data will
                # help us in this regard.
                pass

            plot.decorate_axes(
                ax=ax,
                catalogue=catalogue,
                comment=self.comment,
                legend_loc=self.legend_loc,
                redshift_loc=self.redshift_loc,
                comment_loc=self.comment_loc,
            )

            # Extract y label from matplotlib - autogenerated for us
            self.x_label = ax.get_xlabel()
            self.y_label = ax.get_ylabel()

            if self.x_label_override is not None:
                self.x_label = self.x_label_override
            if self.y_label_override is not None:
                self.y_label = self.y_label_override

        fig.savefig(f"{directory}/{self.filename}.{file_extension}")

        # Delete the figure to cut down on memory consumption.
        close(fig)

        return


class AutoPlotter(object):
    """
    Main autoplotter object; contains all of the VelociraptorPlot objects
    and parsing code to turn the input yaml file into those.
    """

    # Forward declarations
    filename: Union[str, List[str]]
    multiple_yaml_files: bool
    catalogue: VelociraptorCatalogue
    yaml: Dict[str, Union[Dict, str]]
    plots: List[VelociraptorPlot]
    # Directory containing the observational data.
    observational_data_directory: str
    # Whether or not the plots were created successfully.
    created_successfully: List[bool]

    def __init__(
        self,
        filename: Union[str, List[str]],
        observational_data_directory: Union[None, str] = None,
    ) -> None:
        """
        Initialises the AutoPlotter object with the yaml filename(s).

        Parameters
        ----------

        filename: str, List[str]
            Filename(s) of the autoplotter config yaml files you wish to use

        observational_data_directory: str, optional
            Directory containing the observational data to take all paths
            provided under the observational_data section relative to. By
            default this is just ".".
        """

        self.filename = filename

        self.multiple_yaml_files = isinstance(filename, list)
        self.observational_data_directory = (
            observational_data_directory
            if observational_data_directory is not None
            else "."
        )

        self.load_yaml()
        self.parse_yaml()

        return

    def load_yaml(self):
        """
        Loads the yaml data from file.
        """

        if not self.multiple_yaml_files:
            with open(self.filename, "r") as handle:
                self.yaml = safe_load(handle)
        else:
            self.yaml = {}

            for filename in self.filename:
                with open(filename, "r") as handle:
                    self.yaml = {**self.yaml, **safe_load(handle)}

        return

    def parse_yaml(self):
        """
        Parse the contents of the given yaml file into a list of
        VelociraptorPlot instances (self.plots).
        """

        self.plots = [
            VelociraptorPlot(filename, plot, self.observational_data_directory)
            for filename, plot in self.yaml.items()
        ]

        return

    def link_catalogue(self, catalogue: VelociraptorCatalogue):
        """
        Links a catalogue with this object so that the plots
        can actually be created.
        """

        self.catalogue = catalogue

        return

    def create_plots(
        self, directory: str, file_extension: str = "pdf", debug: bool = False
    ):
        """
        Creates and saves the plots in a directory.
        """

        self.file_extension = file_extension

        # Try to create the directory
        if not path.exists(directory):
            mkdir(directory)

        self.created_successfully = []

        for plot in self.plots:
            try:
                plot.make_plot(
                    catalogue=self.catalogue,
                    directory=directory,
                    file_extension=file_extension,
                )
                self.created_successfully.append(True)
            except (AttributeError, ValueError) as e:
                print(f"Unable to create plot {plot.filename} due to exception: {e}.")
                self.created_successfully.append(False)
                if debug:
                    import sys, traceback

                    _, _, exc_traceback = sys.exc_info()
                    print("Traceback:")
                    traceback.print_tb(exc_traceback, limit=10, file=sys.stdout)

        return



"""
Tools to generate various lines from datasets.
"""

import unyt
import numpy as np

from typing import List


def binned_mean_line(
    x: unyt.unyt_array,
    y: unyt.unyt_array,
    x_bins: unyt.unyt_array,
    minimum_in_bin: int = 3,
    return_additional: bool = False,
):
    """
    Gets a mean (y) line, binned in the x direction.

    Parameters
    ----------

    x: unyt.unyt_array
        Horizontal values, to be binned.

    y: unyt.unyt_array
        Vertical values, to have the mean calculated in the x-bins

    x_bins: unyt.unyt_array
        Horizontal bin edges. Must have the same units as x.

    minimum_in_bin: int, optional
        Minimum number of items in a bin to return that bin. If a bin has
        fewer values than this, it is excluded from the return values.
        Default: 3.

    return_additional: bool, optional
        Boolean. If true, makes the function return x and y data points that
        lie in the bins where the number of data points is smaller than
        minimum_in_bin.
        Default: false.


    Returns
    -------

    bin_centers: unyt.unyt_array
        The centers of the bins (taken to be the linear mean of the bin edges).

    means: unyt.unyt_array
        Vertical mean values within the bins.

    standard_deviation: unyt.unyt_array
        Standard deviation within the bins, to be shown as scatter.

    additional_x: unyt.unyt_array, optional
        x data points from the bins where the number of data points is smaller
        than minimum_in_bin

    additional_y: unyt.unyt_array, optional
        y data points from the bins where the number of data points is smaller
        than minimum_in_bin


    Notes
    -----

    The return types are such that you can pass this directly to `plt.errorbar`,
    as follows:

    .. code-block:: python

        plt.errorbar(
            *binned_mean_line(x, y, x_bins, 10)
        )

    """

    assert (
        x.units == x_bins.units
    ), "Please ensure that the x values and bins have the same units."

    hist = np.digitize(x, x_bins)

    means = []
    standard_deviations = []
    centers = []
    additional_x = []
    additional_y = []

    for bin in range(1, len(x_bins)):
        indicies_in_this_bin = hist == bin
        number_of_items_in_bin = indicies_in_this_bin.sum()

        if number_of_items_in_bin >= minimum_in_bin:
            y_values_in_this_bin = y[indicies_in_this_bin].value

            means.append(np.mean(y_values_in_this_bin))
            standard_deviations.append(np.std(y_values_in_this_bin))

            centers.append(0.5 * (x_bins[bin - 1].value + x_bins[bin].value))

        # If the number of data points in the bin is less than minimum_in_bin,
        # collect these data points if needed
        elif number_of_items_in_bin > 0 and return_additional:
            additional_x += list(x[indicies_in_this_bin].value)
            additional_y += list(y[indicies_in_this_bin].value)

    means = unyt.unyt_array(means, units=y.units, name=y.name)
    standard_deviations = unyt.unyt_array(
        standard_deviations, units=y.units, name=f"{y.name} ($sigma$)"
    )
    centers = unyt.unyt_array(centers, units=x.units, name=x.name)
    additional_x = unyt.unyt_array(additional_x, units=x.units, name=x.name)
    additional_y = unyt.unyt_array(additional_y, units=y.units, name=y.name)

    if not return_additional:
        return centers, means, standard_deviations
    else:
        return centers, means, standard_deviations, additional_x, additional_y


def binned_median_line(
    x: unyt.unyt_array,
    y: unyt.unyt_array,
    x_bins: unyt.unyt_array,
    percentiles: List[int] = [16, 84],
    minimum_in_bin: int = 3,
    return_additional: bool = False,
):
    """
    Gets a median (y) line, binned in the x direction.

    Parameters
    ----------

    x: unyt.unyt_array
        Horizontal values, to be binned.

    y: unyt.unyt_array
        Vertical values, to have the median calculated in the x-bins

    x_bins: unyt.unyt_array
        Horizontal bin edges. Must have the same units as x.

    percentiles: List[int], optional
        Percentiles to return as the positive and negative errors. By
        default these are 16 and 84th percentiles.

    minimum_in_bin: int, optional
        Minimum number of items in a bin to return that bin. If a bin has
        fewer values than this, it is excluded from the return values.
        Default: 3.

    return_additional: bool, optional
        Boolean. If true, makes the function return x and y data points that
        lie in the bins where the number of data points is smaller than
        minimum_in_bin.
        Default: false.


    Returns
    -------

    bin_centers: unyt.unyt_array
        The centers of the bins (taken to be the linear mean of the bin edges).

    medians: unyt.unyt_array
        Vertical median values within the bins.

    deviations: unyt.unyt_array
        Deviation from median vertically using the ``percentiles`` defined above.
        This has shape 2xN.

    additional_x: unyt.unyt_array, optional
        x data points from the bins where the number of data points is smaller
        than minimum_in_bin

    additional_y: unyt.unyt_array, optional
        y data points from the bins where the number of data points is smaller
        than minimum_in_bin


    Notes
    -----

    The return types are such that you can pass this directly to `plt.errorbar`,
    as follows:

    .. code-block:: python

        plt.errorbar(
            *binned_median_line(x, y, x_bins, 10)
        )

    """

    assert (
        x.units == x_bins.units
    ), "Please ensure that the x values and bins have the same units."

    hist = np.digitize(x, x_bins)

    medians = []
    deviations = []
    centers = []
    additional_x = []
    additional_y = []

    for bin in range(1, len(x_bins)):
        indicies_in_this_bin = hist == bin
        number_of_items_in_bin = indicies_in_this_bin.sum()

        if number_of_items_in_bin >= minimum_in_bin:
            y_values_in_this_bin = y[indicies_in_this_bin].value

            medians.append(np.median(y_values_in_this_bin))
            deviations.append(np.percentile(y_values_in_this_bin, percentiles))

            centers.append(0.5 * (x_bins[bin - 1].value + x_bins[bin].value))

        # If the number of data points in the bin is less than minimum_in_bin,
        # collect these data points if needed
        elif number_of_items_in_bin > 0 and return_additional:
            additional_x += list(x[indicies_in_this_bin].value)
            additional_y += list(y[indicies_in_this_bin].value)

    medians = unyt.unyt_array(medians, units=y.units, name=y.name)
    # Percentiles actually gives us the values - we want to be able to use
    # matplotlib's errorbar function
    deviations = unyt.unyt_array(
        abs(np.array(deviations).T - medians.value),
        units=y.units,
        name=f"{y.name} {percentiles} percentiles",
    )
    centers = unyt.unyt_array(centers, units=x.units, name=x.name)
    additional_x = unyt.unyt_array(additional_x, units=x.units, name=x.name)
    additional_y = unyt.unyt_array(additional_y, units=y.units, name=y.name)

    if not return_additional:
        return centers, medians, deviations
    else:
        return centers, medians, deviations, additional_x, additional_y



"""
Tools for creating histograms. Uses the same API as mass_functions.
"""

import unyt
import numpy as np

from velociraptor.tools.labels import get_mass_function_label_no_units


def create_histogram_given_bins(
    masses: unyt.unyt_array,
    bins: unyt.unyt_array,
    box_volume: unyt.unyt_quantity,
    minimum_in_bin: int = 1,
):
    """
    Creates a mass function (with equal width bins in log M) for you to plot.

    Parameters
    ----------

    masses: unyt.unyt_array
        The array that you want to create a mass function of (usually this is
        for example halo masses or stellar masses).

    bins: unyt.unyt_array
        The mass bin edges to use.

    unyt.unyt_quantity: box_volume
        The volume of the box such that we can return ``n / volume`` (unused).

    minimum_in_bin: int, optional
        The number of objects in a bin for it to be classed as valid. Bins
        with a number of objects smaller than this are not returned. By default
        this parameter takes a value of 1.


    Returns
    -------

    bin_centers: unyt.unyt_array
        The centers of the bins (taken to be the linear mean of the bin edges).

    histogram: unyt.unyt_array
        The value of the mass function at the bin centers.

    None:
        Final return is None, as there are no errors associated with this.

    """

    bins.convert_to_units(masses.units)

    histogram = unyt.unyt_array(np.histogram(masses, bins)[0], units="dimensionless")
    valid_bins = histogram >= minimum_in_bin

    bin_centers = 0.5 * (bins[1:] + bins[:-1])

    histogram.name = "Number of Haloes"
    bin_centers.name = masses.name

    return bin_centers[valid_bins], histogram[valid_bins], None


def create_histogram(
    masses: unyt.unyt_array,
    lowest_mass: unyt.unyt_quantity,
    highest_mass: unyt.unyt_quantity,
    box_volume: unyt.unyt_quantity,
    n_bins: int = 25,
    minimum_in_bin: int = 1,
    return_bin_edges: bool = False,
):
    """
    Creates a mass function (with equal width bins in log M) for you to plot.

    Parameters
    ----------

    masses: unyt.unyt_array
        The array that you want to create a mass function of (usually this is
        for example halo masses or stellar masses).

    lowest_mass: unyt.unyt_quantity
        the lowest mass edge of the bins
    
    highest_mass: unyt.unyt_quantity
        the highest mass edge of the bins

    bins: unyt.unyt_array
        The mass bin edges to use.

    unyt.unyt_quantity: box_volume
        The volume of the box such that we can return ``n / volume``.

    minimum_in_bin: int, optional
        The number of objects in a bin for it to be classed as valid. Bins
        with a number of objects smaller than this are not returned. By default
        this parameter takes a value of 1.

    return_bin_edges: bool, optional
        Return the bin edges used in the binning process? Default is False.

    Returns
    -------

    bin_centers: unyt.unyt_array
        The centers of the bins (taken to be the linear mean of the bin edges).

    mass_function: unyt.unyt_array
        The value of the mass function at the bin centers.

    None:
        Final return is None, as there are no errors associated with this.

    bin_edges: unyt.unyt_array, optional
        Bin edges that were used in the binning process.

    """

    assert (
        masses.units == lowest_mass.units and lowest_mass.units == highest_mass.units
    ), "Please ensure that all mass quantities have the same units."

    bins = (
        np.logspace(np.log10(lowest_mass), np.log10(highest_mass), n_bins + 1)
        * masses.units
    )

    bin_centers, mass_function, _ = create_histogram_given_bins(
        masses=masses, bins=bins, box_volume=box_volume, minimum_in_bin=minimum_in_bin
    )

    if return_bin_edges:
        return bin_centers, mass_function, None, bins
    else:
        return bin_centers, mass_function, None



"""
Functions that actually plot x against y.
"""

import matplotlib.pyplot as plt
import numpy as np
import unyt

from matplotlib import rcParams

from matplotlib.colors import Normalize, LogNorm
from velociraptor import VelociraptorCatalogue
from velociraptor.autoplotter.objects import VelociraptorLine
from typing import Tuple, Union

import velociraptor.tools as tools


def scatter_x_against_y(
    x: unyt.unyt_array, y: unyt.unyt_quantity
) -> Tuple[plt.Figure, plt.Axes]:
    """
    Creates a scatter of x against y (unyt arrays).
    """

    fig, ax = plt.subplots()

    kwargs = dict(edgecolor="none", zorder=-100)

    # Need to "intelligently" size the markers
    kwargs["s"] = (
        rcParams["lines.markersize"]
        * (6.0 - 5.0 * np.tanh(0.75 * np.log10(x.size) - 3.0))
        / 11.0
    )

    kwargs["alpha"] = (5.5 - 4.5 * np.tanh(0.75 * np.log10(x.size) - 3.0)) / 10.0

    ax.scatter(x.value, y.value, **kwargs)

    return fig, ax


def histogram_x_against_y(
    x: unyt.unyt_array,
    y: unyt.unyt_array,
    x_bins: unyt.unyt_array,
    y_bins: unyt.unyt_array,
) -> Tuple[plt.Figure, plt.Axes]:
    """
    Creates a plot of x against y with a 2d histogram in the background.
    
    Actually uses pcolormesh and the numpy histogram method.
    """

    fig, ax = plt.subplots()

    H, x_bins, y_bins = np.histogram2d(x=x, y=y, bins=[x_bins, y_bins])

    im = ax.pcolormesh(x_bins, y_bins, H.T, norm=LogNorm(), zorder=-100)

    fig.colorbar(im, ax=ax, label="Number of haloes", pad=0.0)

    return fig, ax


def mass_function(
    x: unyt.unyt_array, x_bins: unyt.unyt_array, mass_function: VelociraptorLine
) -> Tuple[plt.Figure, plt.Axes]:
    """
    Creates a plot of x as a mass function, binned with x_bins.
    """

    fig, ax = plt.subplots()

    if mass_function.adaptive_mass_function:
        centers, mass_function, error, *_ = mass_function.output
        ax.errorbar(
            centers, mass_function, yerr=error, xerr=abs(x_bins - centers), fmt=".",
        )
    else:
        centers, mass_function, error, *_ = mass_function.output
        ax.errorbar(centers, mass_function, error)

    return fig, ax


def histogram(
    x: unyt.unyt_array, x_bins: unyt.unyt_array, histogram: VelociraptorLine
) -> Tuple[plt.Figure, plt.Axes]:
    """
    Creates a plot of x as a mass function, binned with x_bins.
    """

    fig, ax = plt.subplots()

    centers, mass_function, *_ = histogram.output

    ax.plot(centers, mass_function)

    return fig, ax


def decorate_axes(
    ax: plt.Axes,
    catalogue: VelociraptorCatalogue,
    comment: Union[str, None] = None,
    legend_loc: str = "upper left",
    redshift_loc: str = "lower right",
    comment_loc: str = "lower left",
) -> None:
    """
    Decorates the axes with information about the redshift and
    scale-factor.
    """

    markerfirst = "right" not in legend_loc

    if len(ax.get_legend_handles_labels()[0]):
        # Only create the legend if we have handles available to plot,
        # otherwise we get a warning from matplotlib printed to the
        # console.
        legend = ax.legend(loc=legend_loc, markerfirst=markerfirst)
        fontsize = legend.get_texts()[0].get_fontsize()
    else:
        fontsize = None

    label_switch = {
        redshift_loc: f"$z={catalogue.z:2.3f}$\n$a={catalogue.a:2.3f}$",
        comment_loc: comment,
    }

    distance_from_edge = 0.025

    for loc, label in label_switch.items():
        if label is not None:
            # First need to parse the 'loc' string
            try:
                va, ha = loc.split(" ")
            except ValueError:
                if loc == "right":
                    ha = "right"
                    va = "center"
                elif loc == "center":
                    ha = "center"
                    va = "center"

            if va == "lower":
                y = distance_from_edge
                va = "bottom"
            elif va == "upper":
                y = 1.0 - distance_from_edge
                va = "top"
            elif va == "center":
                y = 0.5
            else:
                raise AttributeError(
                    f"Unknown location string {loc}. Choose e.g. lower right"
                )

            if ha == "left":
                x = distance_from_edge
            elif ha == "right":
                x = 1.0 - distance_from_edge
            elif ha == "center":
                x = 0.5

            ax.text(
                x,
                y,
                label,
                ha=ha,
                va=va,
                transform=ax.transAxes,
                multialignment=ha,
                fontsize=fontsize,
            )

    return


def get_labels(x: unyt.unyt_array, y: unyt.unyt_array, mass_function: bool) -> None:
    """
    Set the x and y labels for the axes.
    """

    x_label = tools.get_full_label(x)
    y_label = (
        tools.get_mass_function_label(
            mass_function_sub_label="{}", mass_function_units=y.units
        )
        if mass_function
        else tools.get_full_label(y)
    )

    return x_label, y_label



"""
Routines that provide translation of velociraptor quantities into something a
little more human readable, or to internal quantities.
"""

import unyt

from velociraptor.units import VelociraptorUnits


def typo_correct(particle_property_name: str):
    """
    Corrects for any typos in field names that may exist.
    """

    key = {"veldips": "veldisp"}

    if particle_property_name in key.keys():
        return key[particle_property_name]
    else:
        return particle_property_name


def get_aperture_unit(unit_name: str, unit_system: VelociraptorUnits):
    """
    Converts the velociraptor strings to internal velociraptor units 
    from the naming convention in the velociraptor files.
    """

    # Correct any typos
    corrected_name = typo_correct(unit_name).lower()

    key = {
        "sfr": unit_system.star_formation_rate,
        "zmet": unit_system.metallicity,
        "mass": unit_system.mass,
        "npart": unyt.dimensionless,
        "rhalfmass": unit_system.length,
        "veldisp": unit_system.velocity,
        "r": unit_system.length,
        "lx": unit_system.length * unit_system.velocity,
        "ly": unit_system.length * unit_system.velocity,
        "lz": unit_system.length * unit_system.velocity,
    }

    return key.get(corrected_name, None)


def get_particle_property_name_conversion(name: str, ptype: str):
    """
    Takes an internal velociraptor particle property and returns
    a fancier name for use in plot legends. Typically used for the
    complex aperture properties.
    """

    corrected_name = typo_correct(name)

    combined_name = f"{corrected_name}_{ptype}".lower()

    key = {
        "sfr_": "SFR $\dot{\\rho}_*$",
        "sfr_gas": "Gas SFR $\dot{\\rho}_*$",
        "zmet_": "Metallicity $Z$",
        "zmet_gas": "Gas Metallicity $Z_{\\rm g}$",
        "zmet_star": "Star Metallicity $Z_*$",
        "zmet_bh": "Black Hole Metallicity $Z_{\\rm BH}$",
        "mass_": "Mass $M$",
        "mass_gas": "Gas Mass $M_{\\rm g}$",
        "mass_star": "Stellar Mass $M_*$",
        "mass_bh": "Black Hole Mass $M_{\\rm BH}$",
        "mass_interloper": "Mass of Interlopers",
        "npart_": "Number of Particles $N$",
        "npart_gas": "Number of Gas Particles $N_{\\rm g}$",
        "npart_star": "Number of Stellar Particles $N_*$",
        "npart_bh": "Black Hole Mass $N_{\\rm BH}$",
        "npart_interloper": "Number of Interlopers",
        "rhalfmass_": "Half-mass Radius $R_{50}$",
        "rhalfmass_gas": "Gas Half-mass Radius $R_{50, {\\rm g}}$",
        "rhalfmass_star": "Stellar Half-mass Radius $R_{50, *}$",
        "rhalfmass_bh": "Black Hole Half-mass Radius $R_{50, {\\rm BH}}$",
        "r_": "Radius $R_{\\rm SO}$",
        "veldisp_": "Velocity Dispersion $\sigma$",
        "veldisp_gas": "Gas Velocity Dispersion $\sigma_{\\rm g}}$",
        "veldisp_star": "Stellar Velocity Dispersion $\sigma_{*}$",
        "veldisp_bh": "Black Hole Velocity Dispersion $\sigma_{\\rm BH}$",
        "subgridmasses_aperture_total_solar_mass_bh": "Subgrid Black Hole Mass $M_{\\rm BH}$",
    }

    return key.get(combined_name, corrected_name)



"""
Objects for handling and plotting mean and median lines.
"""

from unyt import unyt_quantity, unyt_array
from numpy import logspace, linspace, log10, logical_and
from typing import Dict, Union, Tuple, List
from matplotlib.pyplot import Axes

import velociraptor.tools.lines as lines
from velociraptor.tools.mass_functions import (
    create_mass_function_given_bins,
    create_adaptive_mass_function,
)
from velociraptor.tools.histogram import create_histogram_given_bins
from velociraptor.tools.adaptive import create_adaptive_bins

valid_line_types = [
    "median",
    "mean",
    "mass_function",
    "histogram",
    "adaptive_mass_function",
]


class VelociraptorLine(object):
    """
    A median or mean line and all the information that is
    required for this (e.g. bins, log space, etc.)
    """

    # Forward declarations
    # Actually plot this line?
    plot: bool
    # Is a median, mass function, or a mean line?
    median: bool
    mean: bool
    mass_function: bool
    histogram: bool
    adaptive_mass_function: bool
    # Create bins in logspace?
    log: bool
    # Binning properties
    number_of_bins: int
    start: unyt_quantity
    end: unyt_quantity
    lower: unyt_quantity
    upper: unyt_quantity
    # Use adaptive binning?
    adaptive: bool
    bins: unyt_array = None
    # Scatter can be: "none", "errorbar", or "shaded"
    scatter: str
    # Output: centers, values, scatter, additional_x, additional_y - initialised here
    # to prevent crashes in other code.
    output: Tuple[unyt_array] = (
        unyt_array([]),
        unyt_array([]),
        unyt_array([]),
        unyt_array([]),
        unyt_array([]),
    )

    def __init__(self, line_type: str, line_data: Dict[str, Union[Dict, str]]):
        """
        Initialise a line with data from the yaml file.
        """

        self.line_type = line_type
        self._parse_line_type()

        self.data = line_data
        self._parse_data()

        return

    def _parse_line_type(self):
        """
        Parse the line type to a boolean.
        """

        # TODO: Use centralised metadata for this list.
        for line_type in valid_line_types:
            setattr(self, line_type, self.line_type == line_type)

        return

    def _parse_data(self):
        """
        Parse the line data from the dictionary and set defaults.
        """

        self.plot = bool(self.data.get("plot", True))
        self.log = bool(self.data.get("log", True))
        self.number_of_bins = int(self.data.get("number_of_bins", 25))
        self.scatter = str(self.data.get("scatter", "shaded"))
        self.adaptive = bool(self.data.get("adaptive", False))

        if self.scatter not in ["none", "errorbar", "shaded"]:
            self.scatter = "shaded"

        try:
            self.start = unyt_quantity(
                float(self.data["start"]["value"]), units=self.data["start"]["units"]
            )
        except KeyError:
            self.start = unyt_quantity(0.0)

        try:
            self.end = unyt_quantity(
                float(self.data["end"]["value"]), units=self.data["end"]["units"]
            )
        except KeyError:
            self.end = unyt_quantity(0.0)

        try:
            self.lower = unyt_quantity(
                float(self.data["lower"]["value"]), units=self.data["lower"]["units"]
            )
        except KeyError:
            self.lower = None

        try:
            self.upper = unyt_quantity(
                float(self.data["upper"]["value"]), units=self.data["upper"]["units"]
            )
        except KeyError:
            self.upper = None

        return

    def generate_bins(self, values=None):
        """
        Generates the required bins. If we request adaptive bins, this is a little
        more complicated and requires the values along the horizontal axis.
        """

        if values is not None and self.adaptive:
            self.end.convert_to_units(values.units)
            self.start.convert_to_units(self.end.units)

            bin_centers, bin_edges = create_adaptive_bins(
                values=values,
                lowest_value=self.start,
                highest_value=self.end,
                base_n_bins=self.number_of_bins,
                logarithmic=self.log,
                stretch_final_bin="mass_function" in self.line_type,
            )

            self.bins = bin_edges
        else:
            # Assert these are in the same units just in case
            self.start.convert_to_units(self.end.units)

            if self.log:
                # Need to strip units, unfortunately
                self.bins = unyt_array(
                    logspace(
                        log10(self.start.value),
                        log10(self.end.value),
                        self.number_of_bins,
                    ),
                    units=self.start.units,
                )
            else:
                # Can get away with this one without stripping
                self.bins = linspace(self.start, self.end, self.number_of_bins)

        return

    def create_line(
        self,
        x: unyt_array,
        y: unyt_array,
        box_volume: Union[None, unyt_quantity] = None,
    ):
        """
        Creates the line!

        Parameters
        ----------

        x: unyt_array
            Horizontal axis data
        
        y: unyt_array
            Vertical axis data

        box_volume: Union[None, unyt_quantity]
            Box volume for the simulation, required for mass functions. Should
            have associated volume units. Generally this is given as a comoving
            quantity.

        Returns
        -------

        output: Tuple[unyt_array]
            A five-length (mean, median lines) or three-length (mass_function,
            histogram) tuple of unyt arrays that takes the following form:
            (bin centers, vertical values, vertical scatter, additional_x [optional]
            additional_y [optional]).

        """

        if self.bins is None:
            self.generate_bins(values=x)
        else:
            self.bins.convert_to_units(x.units)

        self.output = None

        masked_x = x
        masked_y = y

        if self.lower is not None:
            self.lower.convert_to_units(y.units)
            mask = masked_y > self.lower
            masked_x = masked_x[mask]
            masked_y = masked_y[mask]

        if self.upper is not None:
            self.upper.convert_to_units(y.units)
            mask = masked_y < self.upper
            masked_x = masked_x[mask]
            masked_y = masked_y[mask]

        if self.median:
            self.output = lines.binned_median_line(
                x=masked_x, y=masked_y, x_bins=self.bins, return_additional=True
            )
        elif self.mean:
            self.output = lines.binned_mean_line(
                x=masked_x, y=masked_y, x_bins=self.bins, return_additional=True
            )
        elif self.mass_function:
            mass_function_output = create_mass_function_given_bins(
                masked_x, self.bins, box_volume=box_volume
            )
            self.output = (
                *mass_function_output,
                unyt_array([], units=mass_function_output[0].units),
                unyt_array([], units=mass_function_output[1].units),
            )
        elif self.histogram:
            histogram_output = create_histogram_given_bins(
                masked_x, self.bins, box_volume=box_volume
            )
            self.output = (
                *histogram_output,
                unyt_array([], units=histogram_output[0].units),
                unyt_array([], units=histogram_output[1].units),
            )
        elif self.adaptive_mass_function:
            *mass_function_output, self.bins = create_adaptive_mass_function(
                masked_x,
                lowest_mass=self.start,
                highest_mass=self.end,
                box_volume=box_volume,
                return_bin_edges=True,
            )
            self.output = (
                *mass_function_output,
                unyt_array([], units=mass_function_output[0].units),
                unyt_array([], units=mass_function_output[1].units),
            )
        else:
            self.output = None

        return self.output

    def plot_line(
        self, ax: Axes, x: unyt_array, y: unyt_array, label: Union[str, None] = None
    ):
        """
        Plot a line using these parameters on some axes, x against y.

        Parameters
        ----------

        ax: Axes
            Matplotlib axes to plot on.

        x: unyt_array
            Horizontal axis data
        
        y: unyt_array
            Vertical axis data

        label: str
            Label associated with this data that will be included in the
            legend.

        Notes
        -----

        If self.scatter is set to "none", this is plotted assuming the scatter
        is zero.
        """

        if not self.plot:
            return

        centers, heights, errors, additional_x, additional_y = self.create_line(
            x=x, y=y
        )

        if self.scatter == "none" or errors is None:
            ax.plot(centers, heights, label=label)
        elif self.scatter == "errorbar":
            ax.errorbar(centers, heights, yerr=errors, label=label)
        elif self.scatter == "errorbar_both":
            ax.errorbar(
                centers,
                heights,
                yerr=errors,
                xerr=abs(self.bins - centers),
                label=label,
                fmt=".",  # Do not plot as a line.
            )
        elif self.scatter == "shaded":
            (line,) = ax.plot(centers, heights, label=label)

            # Deal with different + and -ve errors
            if errors.shape[0]:
                if errors.ndim > 1:
                    down, up = errors
                else:
                    up = errors
                    down = errors
            else:
                up = 0
                down = 0

            ax.fill_between(
                centers,
                heights - down,
                heights + up,
                color=line.get_color(),
                alpha=0.3,
                linewidth=0.0,
            )

        return



"""
Objects for observational data plotting.

Tools for adding in extra (e.g. observational) data to plots.

Includes an object container and helper functions for creating
and reading files.
"""

from unyt import unyt_quantity, unyt_array
from numpy import tanh, log10
from matplotlib.pyplot import Axes
from matplotlib import rcParams

from astropy.units import Quantity
from astropy.cosmology.core import Cosmology
from astropy.cosmology import wCDM, FlatLambdaCDM

import h5py
import json

from typing import Union

# Default z_orders for errorbar points and lines
line_zorder = -5
points_zorder = -6


def save_cosmology(handle: h5py.File, cosmology: Cosmology):
    """
    Save the (astropy) cosmology to a HDF5 dataset.

    Parameters
    ----------

    handle: h5py.File
        h5py file handle to save the cosmology to. This is performed
        by creating a cosmology group and setting attributes.

    cosmology: astropy.cosmology.Cosmology
        The Astropy cosmology instance to save to the HDF5 file. This
        is performed by extracting all of the key variables and saving
        them as either floating point numbers or strings.

    Notes
    -----

    This process can be reversed by using load_cosmology.
    """
    group = handle.create_group("cosmology").attrs

    group.create("H0", cosmology.H0)
    group.create("Om0", cosmology.Om0)
    group.create("Ode0", cosmology.Ode0)
    group.create("Tcmb0", cosmology.Tcmb0)
    group.create("Neff", cosmology.Neff)
    group.create("m_nu", cosmology.m_nu)
    group.create("m_nu_units", str(cosmology.m_nu.unit))
    group.create("Ob0", cosmology.Ob0)
    group.create("name", cosmology.name)

    try:
        group.create("w0", cosmology.w0)
    except:
        # No EoS!
        pass

    return


def load_cosmology(handle: h5py.File):
    """
    Save the (astropy) cosmology to a HDF5 dataset.

    Parameters
    ----------

    handle: h5py.File
        h5py file handle to read the cosmology from.

    Returns
    -------

    astropy.cosmology.Cosmology:
        Astropy cosmology instance extracted from the HDF5 file.
    """

    try:
        group = handle["cosmology"].attrs
    except:
        return None

    try:
        cosmology = wCDM(
            H0=group["H0"],
            Om0=group["Om0"],
            Ode0=group["Ode0"],
            w0=group["w0"],
            Tcmb0=group["Tcmb0"],
            Neff=group["Neff"],
            m_nu=Quantity(group["m_nu"], unit=group["m_nu_units"]),
            Ob0=group["Ob0"],
            name=group["name"],
        )
    except KeyError:
        # No EoS
        cosmology = FlatLambdaCDM(
            H0=group["H0"],
            Om0=group["Om0"],
            Tcmb0=group["Tcmb0"],
            Neff=group["Neff"],
            m_nu=Quantity(group["m_nu"], unit=group["m_nu_units"]),
            Ob0=group["Ob0"],
            name=group["name"],
        )

    return cosmology


class ObservationalData(object):
    """
    Observational data object. Contains routines
    for both writing and reading HDF5 files containing
    the observations, as well as plotting.

    Attributes
    ----------
    name: str
        Name of the observation for users to identifty

    x_units: unyt_quantity
        Units for horizontal axes

    y_units: unyt_quantity
        Units for vertical axes

    x: unyt_array
        Horizontal data points

    y: unyt_array
        Vertical data points

    x_scatter: Union[unyt_array, None]
        Scatter in horizontal direction. Can be None, or an
        unyt_array of shape 1XN (symmetric) or 2XN (non-symmetric)
        such that it can be passed to plt.errorbar easily.

    y_scatter: Union[unyt_array, None]
        Scatter in vertical direction. Can be None, or an
        unyt_array of shape 1XN (symmetric) or 2XN (non-symmetric)
        such that it can be passed to plt.errorbar easily.

    x_comoving: bool
        Whether or not the horizontal values are comoving (True)
        or physical (False)

    y_comoving: bool
        Whether or not the vertical values are comoving (True)
        or physical (False)

    x_description: str
        Default label for horizontal axis (without units), also a
        description of the variable.

    y_description: str
        Default label for horizontal axis (without units), also a
        description of the variable.

    filename: str
        Filename that the data was read from, or was written to.

    comment: str
        A free-text comment describing the data, including e.g.
        which cosmology and IMF it is calibrated to.

    citation: str
        Short citation for data, e.g. Author et al. (Year) (Project),
        such as Baldry et al. (2012) (GAMA)

    bibcode: str
        Bibcode for citation, this can be found on the NASA ADS.

    redshift: float
        Redshift at which the data is collected at. If a range, use
        the mid-point.

    plot_as: Union[str, None]
        Whether the data should be plotted as points (typical for observations)
        or as a line (typical for simulation data). Allowed values:

        + points
        + line

    cosmology: Cosmology
        Astropy cosmology that the data has been corrected to.
    """

    # Data stored in this object
    # name of the observation (to be plotted on axes)
    name: str
    # units for axes
    x_units: unyt_quantity
    y_units: unyt_quantity
    # data for axes
    x: unyt_array
    y: unyt_array
    # scatter
    x_scatter: Union[unyt_array, None]
    y_scatter: Union[unyt_array, None]
    # x and y are comoving?
    x_comoving: bool
    y_comoving: bool
    # x and y labels
    x_description: str
    y_description: str
    # filename to read from or write to
    filename: str
    # free-text comment describing data
    comment: str
    # citation for data
    citation: str
    bibcode: str
    # redshift that the data is at
    redshift: float
    # plot as points, or a line?
    plot_as: Union[str, None] = None
    # the cosmology that this dataset was corrected to
    cosmology: Cosmology

    def __init__(self):
        """
        Initialises the object for observational data. Does nothing as we are
        unsure if we wish to read or write data at this point.
        """

        return

    def load(self, filename: str):
        """
        Loads the observations from file.


        Parameters
        ----------

        filename: str
            The filename to load the data from. Probably should end in
            .hdf5.
        """

        self.filename = filename

        # Load data here.
        self.x = unyt_array.from_hdf5(filename, dataset_name="values", group_name="x")
        self.y = unyt_array.from_hdf5(filename, dataset_name="values", group_name="y")
        self.x_units = self.x.units
        self.y_units = self.y.units

        try:
            self.x_scatter = unyt_array.from_hdf5(
                filename, dataset_name="scatter", group_name="x"
            )
        except KeyError:
            self.x_scatter = None

        try:
            self.y_scatter = unyt_array.from_hdf5(
                filename, dataset_name="scatter", group_name="y"
            )
        except KeyError:
            self.y_scatter = None

        with h5py.File(filename, "r") as handle:
            metadata = handle["metadata"].attrs

            self.comment = metadata["comment"]
            self.name = metadata["name"]
            self.citation = metadata["citation"]
            self.bibcode = metadata["bibcode"]
            self.redshift = metadata["redshift"]
            self.plot_as = metadata["plot_as"]

            self.x_comoving = bool(handle["x"].attrs["comoving"])
            self.y_comoving = bool(handle["y"].attrs["comoving"])
            self.y_description = str(handle["y"].attrs["description"])
            self.x_description = str(handle["x"].attrs["description"])

            self.cosmology = load_cosmology(handle)

        return

    def write(self, filename: str):
        """
        Writes the observations to file.

        Parameters
        ----------

        filename: str
            The filename to write the data to. Probably should end in
            .hdf5.
        """

        self.filename = filename

        # Write data here
        self.x.write_hdf5(filename, dataset_name="values", group_name="x")
        self.y.write_hdf5(filename, dataset_name="values", group_name="y")

        if self.x_scatter is not None:
            self.x_scatter.write_hdf5(filename, dataset_name="scatter", group_name="x")

        if self.y_scatter is not None:
            self.y_scatter.write_hdf5(filename, dataset_name="scatter", group_name="y")

        with h5py.File(filename, "a") as handle:
            metadata = handle.create_group("metadata").attrs

            metadata.create("comment", self.comment)
            metadata.create("name", self.name)
            metadata.create("citation", self.citation)
            metadata.create("bibcode", self.bibcode)
            metadata.create("redshift", self.redshift)
            metadata.create("plot_as", self.plot_as)

            handle["x"].attrs.create("comoving", self.x_comoving)
            handle["y"].attrs.create("comoving", self.y_comoving)
            handle["x"].attrs.create("description", self.x_description)
            handle["y"].attrs.create("description", self.y_description)

            save_cosmology(handle=handle, cosmology=self.cosmology)

        return

    def associate_x(
        self,
        array: unyt_array,
        scatter: Union[unyt_array, None],
        comoving: bool,
        description: str,
    ):
        """
        Associate an x quantity with this observational data instance.

        Parameters
        ----------

        array: unyt_array
            The array of (horizontal) data points, including units.

        scatter: Union[unyt_array, None]
            The array of scatter (1XN or 2XN) in the horizontal
            co-ordinates with associated units.

        comoving: bool
            Whether or not the horizontal values are comoving.

        description: str
            Short description of the data, e.g. Stellar Masses
        """

        self.x = array
        self.x_units = array.units
        self.x_comoving = comoving
        self.x_description = description

        if scatter is not None:
            self.x_scatter = scatter.to(self.x_units)
        else:
            self.x_scatter = None

        return

    def associate_y(
        self,
        array: unyt_array,
        scatter: Union[unyt_array, None],
        comoving: bool,
        description: str,
    ):
        """
        Associate an y quantity with this observational data instance.

        Parameters
        ----------

        array: unyt_array
            The array of (vertical) data points, including units.

        scatter: Union[unyt_array, None]
            The array of scatter (1XN or 2XN) in the vertical
            co-ordinates with associated units.

        comoving: bool
            Whether or not the vertical values are comoving.

        description: str
            Short description of the data, e.g. Stellar Masses
        """

        self.y = array
        self.y_units = array.units
        self.y_comoving = comoving
        self.y_description = description

        if scatter is not None:
            self.y_scatter = scatter.to(self.y_units)
        else:
            self.y_scatter = None

        return

    def associate_citation(self, citation: str, bibcode: str):
        """
        Associate a citation with this observational data instance.

        Parameters
        ----------

        citation: str
            Short citation, formatted as follows: Author et al. (Year) (Project),
            e.g. Baldry et al. (2012) (GAMA)

        bibcode: str
            Bibcode for the paper the data was extracted from, available
            from the NASA ADS or publisher. E.g. 2012MNRAS.421..621B
        """

        self.citation = citation
        self.bibcode = bibcode

        return

    def associate_name(self, name: str):
        """
        Associate a name with this observational data instance.
        
        Parameters
        ----------

        name: str
            Short name to describe the dataset.
        """

        self.name = name

        return

    def associate_comment(self, comment: str):
        """
        Associate a comment with this observational data instance.

        Parameters
        ----------

        comment: str
            A free-text comment describing the data, including e.g.
            which cosmology and IMF it is calibrated to.
        """

        self.comment = comment

        return

    def associate_redshift(self, redshift: float):
        """
        Associate the redshift that the observations were taken at
        with this observational data instance.

        Parameters
        ----------

        redshift: float
            Redshift at which the data is collected at. If a range, use
            the mid-point.

        """

        self.redshift = redshift

        return

    def associate_plot_as(self, plot_as: str):
        """
        Associate the 'plot_as' field - this should either be line
        or points.

        Parameters
        ----------

        plot_as: str
            Either points or line
        """

        if plot_as not in ["line", "points"]:
            raise Exception("Please supply plot_as as either points or line.")

        self.plot_as = plot_as

        return

    def associate_cosmology(self, cosmology: Cosmology):
        """
        Associate a cosmology with this dataset that it has been corrected for.
        This should be an astropy cosmology instance.

        Parameters
        ----------

        cosmology: astropy.cosmology.Cosmology
            Astropy cosmology instance describing what cosmology the data has
            been corrected to.
        """

        self.cosmology = cosmology

        return

    def plot_on_axes(self, axes: Axes, errorbar_kwargs: Union[dict, None] = None):
        """
        Plot this set of observational data as an errorbar().

        Parameters
        ----------

        axes: plt.Axes
            The matplotlib axes to plot the data on. This will either
            plot the data as a line or a set of errorbar points, with
            the short citation (self.citation) being included in the
            legend automatically.

        errorbar_kwargs: dict
            Optional keyword arguments to pass to plt.errorbar.
        """

        # Do this because dictionaries are mutable
        if errorbar_kwargs is not None:
            kwargs = errorbar_kwargs
        else:
            kwargs = {}

        # Ensure correct units throughout, in case somebody changed them
        if self.x_scatter is not None:
            self.x_scatter.convert_to_units(self.x.units)

        if self.y_scatter is not None:
            self.y_scatter.convert_to_units(self.y.units)

        if self.plot_as == "points":
            kwargs["linestyle"] = "none"
            kwargs["marker"] = "."
            kwargs["zorder"] = points_zorder

            # Need to "intelligently" size the markers
            kwargs["markersize"] = (
                rcParams["lines.markersize"]
                * (1.5 - tanh(2.0 * log10(len(self.x)) - 4.0))
                / 2.5
            )

            kwargs["alpha"] = (3.0 - tanh(2.0 * log10(len(self.x)) - 4.0)) / 4.0

            # Looks weird if errorbars are present
            if self.y_scatter is None:
                kwargs["markerfacecolor"] = "none"

            if len(self.x) > 1000:
                kwargs["rasterize"] = True
        elif self.plot_as == "line":
            kwargs["zorder"] = line_zorder

        # Make both the data name and redshift appear in the legend
        data_label = f"{self.citation} ($z={self.redshift:.1f}$)"

        axes.errorbar(
            self.x,
            self.y,
            yerr=self.y_scatter,
            xerr=self.x_scatter,
            **kwargs,
            label=data_label
        )

        return



"""
Contains a class for collecting and writing metadata about velociraptor
plots.
"""

from velociraptor.autoplotter.objects import (
    VelociraptorPlot,
    AutoPlotter,
    valid_line_types,
    valid_plot_types,
)

from velociraptor.autoplotter.lines import VelociraptorLine

from typing import List, Dict

import yaml
from unyt import unyt_array


class VelociraptorLineMetadata(object):
    """
    Individual mnetadat for a given velociraptor line.
    """

    # Forward declarations

    def __init__(self, line: VelociraptorLine):
        """
        Metadata for a given line, including the values of
        that line.
        """

        self.line = line
        self._parse_line()

        return

    def _parse_line(self):
        """
        Parse the line to x and y arrays.
        """

        (
            self.centers,
            self.values,
            self.scatter,
            self.additional_x,
            self.additional_y,
        ) = self.line.output
        self.line_type = self.line.line_type

        return

    def to_dict(self):
        """
        Convert the contents of this into a dictionary (not using
        __dict__ as that is used for hashing and I don't want to 
        mess with that here)
        """

        # Need to convert numeric values to lists and units to strings
        # so that they are written out nicely with the yaml library,
        # otherwise it is not human readable.
        output = dict(
            centers=self.centers.value.tolist(),
            centers_units=str(self.centers.units),
            values=self.values.value.tolist(),
            values_units=str(self.values.units),
            line_type=self.line_type,
            additional_points_x=self.additional_x.value.tolist(),
            additional_points_x_units=str(self.additional_x.units),
            additional_points_y=self.additional_y.value.tolist(),
            additional_points_y_units=str(self.additional_y.units),
            bins_x=self.line.bins.value.tolist(),
            bins_x_units=str(self.line.bins.units),
        )

        try:
            output["scatter"] = self.scatter.value.tolist()
            output["scatter_units"] = str(self.scatter.units)
        except (TypeError, AttributeError):
            output["scatter"] = [0.0] * len(self.centers)
            output["scatter_units"] = "dimensionless"

        return output


class VelociraptorPlotMetadata(object):
    """
    Individual metadata for a given velociraptor plot.
    """

    # Forward declarations
    # Plot associated with me
    plot: VelociraptorPlot
    # Metadata dictionary
    metadata: Dict[str, str]
    # Should we write lines?
    write_lines: bool
    lines: List[VelociraptorLineMetadata]
    # Fancy name for the plot
    title: str
    # Section plot lives in
    section: str
    # Caption for the plot
    caption: str
    # Should this be shown in a webpage or just created?
    show_on_webpage: bool

    def __init__(self, plot: VelociraptorPlot):
        """
        Give me the plot, and I will do all the work!
        """

        self.plot = plot
        self._parse_metadata()

        return

    def _parse_metadata_section_from_file(self):
        """
        Parses metadata: title.
        """

        self.title = self.metadata.get("title", "")
        self.caption = self.metadata.get("caption", "")
        self.section = self.metadata.get("section", "")
        self.show_on_webpage = self.metadata.get("show_on_webpage", True)
        self.filename = self.plot.filename

        return

    def _parse_line_write(self):
        """
        Parses metadata: write_lines, which determines if we
        will write the lines associated with this velociraptor plot
        out to file too.
        """

        self.write_lines = bool(self.metadata.get("write_lines", True))

        return

    def _parse_metadata(self):
        """
        Parses the metadata from the metadata: header in the
        plot.
        """

        self.metadata = self.plot.data.get("metadata", {})

        self._parse_metadata_section_from_file()
        self._parse_line_write()
        self._parse_lines()
        self._parse_labels()

        return

    def _parse_lines(self):
        """
        Parses the information in the current VelociraptorPlot's lines
        instances to VelociraptorLineMetadata.
        """

        self.lines = []

        for line_type in valid_line_types:
            try:
                line = getattr(self.plot, f"{line_type}_line")
            except AttributeError:
                # This line must not be present!
                continue

            if line:
                self.lines.append(VelociraptorLineMetadata(line=line))

        return

    def _parse_labels(self):
        """
        Parses the labels (without units).
        """

        self.x_quantity = self.plot.x
        self.y_quantity = self.plot.y

        self.x_label = self.plot.x_label
        self.y_label = self.plot.y_label

        return

    def to_dict(self):
        """
        Converts the contents of this into a dictionary, avoiding
        the use of __dict__.
        """

        output = dict(
            title=self.title,
            section=self.section,
            caption=self.caption,
            show_on_webpage=self.show_on_webpage,
            filename=self.filename,
            x_quantity=self.x_quantity,
            y_quantity=self.y_quantity,
            x_label=self.x_label,
            y_label=self.y_label,
        )

        if self.write_lines:
            output["lines"] = {line.line_type: line.to_dict() for line in self.lines}

        return output


class AutoPlotterMetadata(object):
    """
    Contains metadata about the autoplotter, and in particular
    about the plots - for instance the values actually retrieved
    from the lines.
    """

    # Individual metadata
    plots: List[VelociraptorPlot]
    metadata: List[VelociraptorPlotMetadata]

    def __init__(self, auto_plotter: AutoPlotter):
        """
        Give me an instance of AutoPlotter and I will do the rest of the
        work.
        """

        self.auto_plotter = auto_plotter
        self.plots = auto_plotter.plots
        self.file_extension = auto_plotter.file_extension

        self._generate_plotter_metadata()

        return

    def _generate_plotter_metadata(self):
        """
        Generate all of the individual plot metadata from the
        instances of VelociraptorPlot.
        """

        self.metadata = [VelociraptorPlotMetadata(plot=plot) for plot in self.plots]

        return

    def write_metadata(self, filename: str):
        """
        Writes the metadata out to a (yaml) file.
        """

        metadata = {plot.filename: plot.to_dict() for plot in self.metadata}

        try:
            metadata["metadata"] = dict(
                redshift=float(self.auto_plotter.catalogue.units.redshift),
                scale_factor=float(self.auto_plotter.catalogue.units.scale_factor),
            )
        except:
            pass

        with open(filename, "w") as handle:
            yaml.dump(metadata, stream=handle)

        return



"""
Manually defined exceptions.
"""


class RegistrationDoesNotMatchError(Exception):
    """
    Raised when our registration function does not match the current
    variable name passed to it.
    """

    def __init__(self, message=""):

        # Call the base class constructor with the parameters it needs
        super().__init__(message)


class AutoPlotterError(Exception):
    def __init__(self, message):
        self.message = message



"""
Tools for creating adaptive bins based on an input array.
"""

import numpy as np
import unyt


def create_adaptive_bins(
    values: unyt.unyt_array,
    lowest_value: unyt.unyt_quantity,
    highest_value: unyt.unyt_quantity,
    base_n_bins: int = 25,
    minimum_in_bin: int = 3,
    logarithmic: bool = True,
    stretch_final_bin: bool = False,
):
    """
    Creates a set of adaptive bins based on the input values.

    Parameters
    ----------

    values: unyt.unyt_array
        The array that you want to create a mass function of (usually this is
        for example halo masses or stellar masses).

    lowest_value: unyt.unyt_quantity
        the lowest value edge of the bins
    
    highest_value: unyt.unyt_quantity
        the highest value edge of the bins

    base_n_bins: unyt.unyt_array, optional
        The number of equal width bins across the range to use in the case
        where no adaptive sampling is required. This returns the minimal allowed
        bin width. Default: 25.

    minimum_in_bin: int, optional
        The number of objects in a bin for it to be classed as valid. Bins
        with a number of objects smaller than this are not returned. Bins are
        stretched so that they have at least this many items in them.
        Default: 3.

    logarithmic: bool, optional
        Whether or not to use logarithmically spaced bins. Default: True.

    stretch_final_bin: bool, optional
        Stretch the final bin to include all values. If ``False``, some values
        may fall outside of all of the bins.

    Returns
    -------

    bin_centers: unyt.unyt_array
        The centers of the bins (taken to be the median of the items in the bin).

    bin_edges: unyt.unyt_array, optional
        Bin edges that were used in the binning process.

    """
    assert (
        values.units == lowest_value.units and lowest_value.units == highest_value.units
    ), "Please ensure that all value quantities have the same units."

    # We must do this and not modify the data internally as it could be used for
    # plotting elsewhere.
    sorted_values = np.sort(values)

    mask = np.logical_and(
        sorted_values >= lowest_value, sorted_values <= highest_value,
    )

    if logarithmic:
        sorted_values = np.log10(sorted_values[mask].value)

        minimal_bin_width = (
            np.log10(highest_value.value) - np.log10(lowest_value.value)
        ) / base_n_bins
    else:
        minimal_bin_width = ((highest_value - lowest_value) / base_n_bins).value
        sorted_values = sorted_values.value

    number_in_bin = []
    bin_edges_left = [
        np.log10(lowest_value.value) if logarithmic else lowest_value.value
    ]
    bin_edges_right = []
    bin_medians = []

    current_edge_left = bin_edges_left[0]
    current_lower_index = 0
    current_bin_count = 0

    for index, value in enumerate(sorted_values):
        current_bin_count += 1

        if value < current_edge_left + minimal_bin_width:
            # Nothing to do, we're just filling the bin
            continue
        elif current_bin_count > minimum_in_bin:
            # Let's just end this here
            bin_medians.append(
                np.median(sorted_values[current_lower_index : index + 1])
            )

            # The new bin edge lives half way in between our current value and the
            # next value, if it exists.
            try:
                new_edge = 0.5 * (sorted_values[index + 1] + value)
            except IndexError:
                # This case is where `value` is the last item in the array.
                new_edge = value

            bin_edges_right.append(new_edge)
            bin_edges_left.append(new_edge)
            number_in_bin.append(current_bin_count)

            # Reset for the lads
            current_edge_left = new_edge
            current_lower_index = index
            current_bin_count = 0
        else:
            # Nothing we can do! Let's just keep going and
            # hope we find more matches.
            pass

    if stretch_final_bin:
        # Clean up any left-overs.
        if current_bin_count > 1:
            # We can create a novel bin!
            bin_edges_right.append(value)
            bin_medians.append(
                np.median(sorted_values[current_lower_index : index + 1])
            )
            number_in_bin.append(current_bin_count)
        elif current_bin_count == 1:
            # Extend the previous bin (otherwise error is consistent with zero)
            bin_edges_right[-1] = value
            number_in_bin[-1] += 1
            bin_medians[-1] = np.median(sorted_values[-number_in_bin[-1] :])
            # We don't need the next bin now.
            bin_edges_left = bin_edges_left[:-1]
        else:
            # This bin doesn't exist anyway, boo...
            bin_edges_left = bin_edges_left[:-1]
    else:
        bin_edges_left = bin_edges_left[:-1]

    if logarithmic:
        bin_centers = unyt.unyt_array(
            [10 ** x for x in bin_medians], units=values.units, name=values.name
        )

        bin_edges = unyt.unyt_array(
            10 ** np.array([*bin_edges_left, bin_edges_right[-1]]), units=values.units,
        )
    else:
        bin_centers = unyt.unyt_array(bin_medians, units=values.units, name=values.name)

        bin_edges = unyt.unyt_array(
            np.array([*bin_edges_left, bin_edges_right[-1]]), units=values.units
        )

    return bin_centers, bin_edges

